# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Branch data types"""

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.constants import empty
from power_grid_model_ds._core.model.dtypes.id import Id


class Branch(Id):
    """Branch data type"""

    from_node: NDArray[np.int32]  # node id (from-side)
    to_node: NDArray[np.int32]  # node id (to-side)
    from_status: NDArray[np.int8]  # 1 = closed, 0 = open
    to_status: NDArray[np.int8]  # 1 = closed, 0 = open
    feeder_branch_id: NDArray[np.int32]  # branch id of the feeding branch
    feeder_node_id: NDArray[np.int32]  # node id of the feeding node
    is_feeder: NDArray[np.bool_]  # whether or not this branch is from the substation

    _defaults = {
        "feeder_branch_id": empty,
        "feeder_node_id": empty,
        "is_feeder": False,
    }


class Link(Branch):
    """Link data type"""


class Line(Branch):
    """Line data type"""

    r1: NDArray[np.float64]  # serial resistance
    x1: NDArray[np.float64]  # serial reactance
    c1: NDArray[np.float64]  # shunt capacitance
    tan1: NDArray[np.float64]  # shunt loss factor
    i_n: NDArray[np.float64]  # rated current


class GenericBranch(Branch):
    """GenericBranch data type (generic_branch in power-grid-model)

    Off-nominal ratio k and phase shift theta are modelled explicitly. Rated power sn optional.
    The impedance (r1, x1) and admittance (g1, b1) are given wrt the to-side.
    """

    r1: NDArray[np.float64]  # positive-sequence resistance
    x1: NDArray[np.float64]  # positive-sequence reactance
    g1: NDArray[np.float64]  # positive-sequence conductance
    b1: NDArray[np.float64]  # positive-sequence susceptance


class Transformer(Branch):
    """Transformer data type"""

    u1: NDArray[np.float64]  # rated voltage (from-side)
    u2: NDArray[np.float64]  # rated voltage (to-side)
    sn: NDArray[np.float64]  # rated power
    tap_size: NDArray[np.float64]  # size of each tap of the tap changer
    uk: NDArray[np.float64]  # relative short circuit voltage
    pk: NDArray[np.float64]  # short circuit loss
    i0: NDArray[np.float64]  # relative no-load current
    p0: NDArray[np.float64]  # no-load loss
    winding_from: NDArray[np.int8]  # winding type (from-side)
    winding_to: NDArray[np.int8]  # winding type (to-side)
    clock: NDArray[np.int8]  # clock number of phase shift
    tap_side: NDArray[np.int8]  # side of tap changer
    tap_pos: NDArray[np.int8]  # current position of tap changer
    tap_min: NDArray[np.int8]  # position of tap changer at minimum voltage
    tap_max: NDArray[np.int8]  # position of tap changer at maximum voltage
    tap_nom: NDArray[np.int8]  # nominal position of tap changer


class Branch3(Id):
    """Branch3 data type"""

    node_1: NDArray[np.int32]
    node_2: NDArray[np.int32]
    node_3: NDArray[np.int32]
    status_1: NDArray[np.int8]
    status_2: NDArray[np.int8]
    status_3: NDArray[np.int8]


class ThreeWindingTransformer(Branch3):
    """ThreeWindingTransformer data type"""

    u1: NDArray[np.float64]
    u2: NDArray[np.float64]
    u3: NDArray[np.float64]
    sn_1: NDArray[np.float64]
    sn_2: NDArray[np.float64]
    sn_3: NDArray[np.float64]
    uk_12: NDArray[np.float64]
    uk_13: NDArray[np.float64]
    uk_23: NDArray[np.float64]
    pk_12: NDArray[np.float64]
    pk_13: NDArray[np.float64]
    pk_23: NDArray[np.float64]
    i0: NDArray[np.float64]
    p0: NDArray[np.float64]
    winding_1: NDArray[np.int8]
    winding_2: NDArray[np.int8]
    winding_3: NDArray[np.int8]
    clock_12: NDArray[np.int8]
    clock_13: NDArray[np.int8]
    tap_side: NDArray[np.int8]
    tap_pos: NDArray[np.int8]
    tap_min: NDArray[np.int8]
    tap_max: NDArray[np.int8]
    tap_nom: NDArray[np.int8]
    tap_size: NDArray[np.float64]
    uk_12_min: NDArray[np.float64]
    uk_13_min: NDArray[np.float64]
    uk_23_min: NDArray[np.float64]
    pk_12_min: NDArray[np.float64]
    pk_13_min: NDArray[np.float64]
    pk_23_min: NDArray[np.float64]
    uk_12_max: NDArray[np.float64]
    uk_13_max: NDArray[np.float64]
    uk_23_max: NDArray[np.float64]
    pk_12_max: NDArray[np.float64]
    pk_13_max: NDArray[np.float64]
    pk_23_max: NDArray[np.float64]


class AsymLine(Branch):
    """Asymmetric Line data type (asym_line in power-grid-model)

    Supports 3 or 4 phase (with neutral) resistance / reactance matrices and optional capacitance matrix
    or sequence capacitances c0/c1. If c_* matrix is omitted, c0 & c1 may be specified instead.
    Only include fields; validation logic handled elsewhere (not implemented here yet).
    """

    # Resistance matrix entries (series)
    r_aa: NDArray[np.float64]
    r_ba: NDArray[np.float64]
    r_bb: NDArray[np.float64]
    r_ca: NDArray[np.float64]
    r_cb: NDArray[np.float64]
    r_cc: NDArray[np.float64]

    # Reactance matrix entries (series)
    x_aa: NDArray[np.float64]
    x_ba: NDArray[np.float64]
    x_bb: NDArray[np.float64]
    x_ca: NDArray[np.float64]
    x_cb: NDArray[np.float64]
    x_cc: NDArray[np.float64]

    # Alternative sequence capacitances
    c0: NDArray[np.float64]
    c1: NDArray[np.float64]

    i_n: NDArray[np.float64]  # rated current



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING

from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.utils.pickle import get_pickle_path, load_from_pickle, save_to_pickle
from power_grid_model_ds._core.utils.zip import file2gzip

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid


def load_grid_from_pickle(grid_class: type["Grid"], cache_path: Path, load_graphs: bool = True):
    """See Grid.from_cache()"""
    pickle_path = get_pickle_path(cache_path)
    grid = load_from_pickle(path=pickle_path)
    if not isinstance(grid, grid_class):
        raise TypeError(f"{pickle_path.name} is not a valid {grid_class.__name__} cache.")

    if load_graphs:
        grid.graphs = GraphContainer.from_arrays(grid)
    return grid


def save_grid_to_pickle(grid: "Grid", cache_dir: Path, cache_name: str, compress: bool = True):
    """See Grid.cache()"""
    tmp_graphs = copy(grid.graphs)
    grid.graphs = None  # noqa
    cache_dir.mkdir(parents=True, exist_ok=True)

    pickle_path = cache_dir / f"{cache_name}.pickle"
    save_to_pickle(path=pickle_path, python_object=grid)

    if compress:
        gzip_path = file2gzip(pickle_path)
        pickle_path.unlink()
        return gzip_path

    grid.graphs = tmp_graphs
    return pickle_path



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

try:
    from power_grid_model_ds._core.visualizer.app import visualize
except ImportError as error:
    raise ImportError(
        "Missing dependencies for visualizer: install with 'pip install power-grid-model-ds[visualizer]'"
    ) from error

__all__ = ["visualize"]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Arrays"""

from typing import Literal

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.fancypy import concatenate
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.dtypes.appliances import Source, SymGen, SymLoad
from power_grid_model_ds._core.model.dtypes.branches import (
    AsymLine,
    Branch,
    Branch3,
    GenericBranch,
    Line,
    Link,
    ThreeWindingTransformer,
    Transformer,
)
from power_grid_model_ds._core.model.dtypes.id import Id
from power_grid_model_ds._core.model.dtypes.nodes import Node
from power_grid_model_ds._core.model.dtypes.regulators import TransformerTapRegulator
from power_grid_model_ds._core.model.dtypes.sensors import (
    AsymCurrentSensor,
    AsymPowerSensor,
    AsymVoltageSensor,
    SymCurrentSensor,
    SymPowerSensor,
    SymVoltageSensor,
)

# pylint: disable=missing-class-docstring


class IdArray(Id, FancyArray):
    pass


class SymLoadArray(IdArray, SymLoad):
    pass


class SymGenArray(IdArray, SymGen):
    pass


class SourceArray(IdArray, Source):
    pass


class NodeArray(IdArray, Node):
    pass


class BranchArray(IdArray, Branch):
    @property
    def node_ids(self):
        """Return both from_node and to_node in one array"""
        return np.concatenate([self.data["from_node"], self.data["to_node"]])

    @property
    def is_active(self) -> NDArray[np.bool_]:
        """Returns boolean whether branch is closed at both ends"""
        return np.logical_and(self.from_status == 1, self.to_status == 1)

    def filter_parallel(self, n_parallel: int, mode: Literal["eq", "neq"]) -> "BranchArray":
        """Return branches that have n_parallel connections.

        Args:
            branches: BranchArray.
            n_parallel: the number of connections between the same nodes
            mode: mode of comparison. "eq" (equal) or "neq" (non-equal).

        Returns:
            - when n_parallel is 1 and mode is 'eq', the function returns branches that are not parallel.
            - when n_parallel is 1 and mode is 'neq', the function returns branches that are parallel.
        """
        _, index, counts = np.unique(self[["from_node", "to_node"]], return_counts=True, return_index=True)

        match mode:
            case "eq":
                counts_mask = counts == n_parallel
            case "neq":
                counts_mask = counts != n_parallel
            case _:
                raise ValueError(f"mode {mode} not supported")

        if mode == "eq" and n_parallel == 1:
            return self[index][counts_mask]
        filtered_branches = self[index][counts_mask]
        return self.filter(from_node=filtered_branches.from_node, to_node=filtered_branches.to_node)


class LinkArray(Link, BranchArray):
    pass


class LineArray(Line, BranchArray):
    pass


class TransformerArray(Transformer, BranchArray):
    pass


class GenericBranchArray(GenericBranch, BranchArray):
    pass


class AsymLineArray(AsymLine, BranchArray):
    pass


class Branch3Array(IdArray, Branch3):
    def as_branches(self) -> BranchArray:
        """Convert Branch3Array to BranchArray."""
        branches_1_2 = BranchArray.empty(self.size)
        branches_1_2.from_node = self.node_1
        branches_1_2.to_node = self.node_2
        branches_1_2.from_status = self.status_1
        branches_1_2.to_status = self.status_2

        branches_1_3 = BranchArray.empty(self.size)
        branches_1_3.from_node = self.node_1
        branches_1_3.to_node = self.node_3
        branches_1_3.from_status = self.status_1
        branches_1_3.to_status = self.status_3

        branches_2_3 = BranchArray.empty(self.size)
        branches_2_3.from_node = self.node_2
        branches_2_3.to_node = self.node_3
        branches_2_3.from_status = self.status_2
        branches_2_3.to_status = self.status_3
        return concatenate(branches_1_2, branches_1_3, branches_2_3)


class ThreeWindingTransformerArray(Branch3Array, ThreeWindingTransformer):
    pass


class TransformerTapRegulatorArray(IdArray, TransformerTapRegulator):
    pass


class SymPowerSensorArray(IdArray, SymPowerSensor):
    pass


class SymVoltageSensorArray(IdArray, SymVoltageSensor):
    pass


class SymCurrentSensorArray(IdArray, SymCurrentSensor):
    pass


class AsymPowerSensorArray(IdArray, AsymPowerSensor):
    pass


class AsymVoltageSensorArray(IdArray, AsymVoltageSensor):
    pass


class AsymCurrentSensorArray(IdArray, AsymCurrentSensor):
    pass



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash
from dash import Input, Output, callback

from power_grid_model_ds._core.visualizer.layout.header import CONFIG_DIV, LEGENDA_DIV, SEARCH_DIV


@callback(
    Output("header-right-col", "children"),
    [
        Input("btn-legend", "n_clicks"),
        Input("btn-search", "n_clicks"),
        Input("btn-config", "n_clicks"),
    ],
)
def update_right_col(_btn1, _btn2, _btn3):
    """Update the right column content based on the button clicked."""
    ctx = dash.callback_context
    if not ctx.triggered:
        return LEGENDA_DIV
    button_id = ctx.triggered[0]["prop_id"].split(".")[0]
    button_map = {
        "btn-legend": LEGENDA_DIV,
        "btn-search": SEARCH_DIV,
        "btn-config": CONFIG_DIV,
    }
    return button_map.get(button_id, "Right Column")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import TYPE_CHECKING

import numpy as np

from power_grid_model_ds._core.model.arrays import BranchArray
from power_grid_model_ds._core.model.enums.nodes import NodeType

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid


def set_feeder_ids(grid: "Grid"):
    """Determines all feeder groups in the network and gives them a feeder identifier

    feeder_branch_id := All assets being connected through the same feeding branch(es) on the substation
    feeder_node_id := All assets connected to the same feeder node (substation)

    Example:
    Nodes Topology
    (SUBSTATION) 101 --- 102 --- 103 -|- 104 --- 105 --- 101 (SUBSTATION)
                         {-}
                         106

    Branches Topology:
    (SUBSTATION) *** 201 *** 202 *** 203 *** 601 *** 204 *** (SUBSTATION)
                         301
                         ***

         Substation ID | Feeder ID
    101 | -1           | -1
    102 | 101          | 201
    103 | 101          | 201
    104 | 101          | 204
    105 | 101          | 204
    106 | 101          | 201
    201 | 101          | 201
    202 | 101          | 201
    203 | 101          | -1
    204 | 101          | 204
    301 | 101          | 201
    601 | 101          | 204
    """
    _set_is_feeder(grid=grid)
    _reset_feeder_ids(grid)
    feeder_node_ids = grid.node.filter(node_type=NodeType.SUBSTATION_NODE)["id"]
    with grid.graphs.active_graph.tmp_remove_nodes(feeder_node_ids.tolist()):
        components = grid.graphs.active_graph.get_components()
    for component_node_ids in components:
        component_branches = _get_active_component_branches(grid, component_node_ids)

        feeder_branch = _get_feeder_branch(component_branches)

        if feeder_branch.size == 0:
            continue  # early exit

        feeder_node_id = _get_feeder_node_id(feeder_branch, feeder_node_ids)

        for array in grid.branch_arrays:
            array.update_by_id(
                component_branches.id,
                feeder_branch_id=feeder_branch.id.item(),
                feeder_node_id=feeder_node_id,
                allow_missing=True,
            )

        grid.node.update_by_id(
            component_node_ids,
            feeder_branch_id=feeder_branch.id.item(),
            feeder_node_id=feeder_node_id,
            allow_missing=True,
        )


def _set_is_feeder(grid: "Grid") -> None:
    "Set the is_feeder property for all branches in the network."
    feeder_node_ids = grid.node.filter(node_type=NodeType.SUBSTATION_NODE).id
    array: BranchArray
    for array in [grid.link, grid.line, grid.transformer]:
        array.is_feeder = np.logical_xor(
            np.isin(array.from_node, feeder_node_ids), np.isin(array.to_node, feeder_node_ids)
        )


def _reset_feeder_ids(grid: "Grid"):
    # Resets all feeder ids to EMPTY_ID
    for array in grid.branch_arrays:
        array.set_empty("feeder_branch_id")
        array.set_empty("feeder_node_id")

    grid.node.set_empty("feeder_branch_id")
    grid.node.set_empty("feeder_node_id")


def _get_active_component_branches(grid: "Grid", component_node_ids: list[int]) -> BranchArray:
    # a component is a set of actively connected nodes (ids)
    # returns all active branches in the component

    branches = grid.branches
    branches_in_component = branches.filter(from_node=component_node_ids, to_node=component_node_ids, mode_="OR")
    active_branches = branches_in_component.filter(from_status=1, to_status=1)
    return active_branches


def _get_feeder_branch(component_branches: BranchArray) -> BranchArray:
    feeder_branches = component_branches.filter(is_feeder=True)

    if feeder_branches.size == 1:
        return feeder_branches

    if feeder_branches.size > 1:
        # Cannot point to multiple branches, so just pick the first one
        return feeder_branches[0]

    return BranchArray()


def _get_feeder_node_id(feeder_branch: BranchArray, feeder_node_ids: np.ndarray) -> int:
    # intersect to retrieve the feeder node id
    feeder_node_id = np.intersect1d(feeder_branch.node_ids, feeder_node_ids)
    return feeder_node_id.item()



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import html

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS

_BOOTSTRAP_ARROW_ICON_CLASS = "fas fa-arrow-right-long"
_MARGIN = "0 10px"

LEGENDA_STYLE = {"margin": _MARGIN, "text-shadow": "0 0 5px #000", "justify-content": "center"}

_FONT_SIZE = "2.5em"

NODE_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["node"]}
_SUBSTATION_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["substation_node"]}
_LINE_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["line"]}
_LINK_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["link"]}
_TRANSFORMER_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["transformer"]}
_GENERIC_BRANCH_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["generic_branch"]}
_ASYM_LINE_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["asym_line"]}
_OPEN_BRANCH_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["open_branch"]}
LEGENDA_ELEMENTS = [
    html.I(className="fas fa-circle", id="node-icon", style=NODE_ICON_STYLE),
    dbc.Tooltip("Node", target="node-icon", placement="bottom"),
    html.I(className="fas fa-diamond", id="substation-icon", style=_SUBSTATION_ICON_STYLE),
    dbc.Tooltip("Substation", target="substation-icon", placement="bottom"),
    html.I(className=_BOOTSTRAP_ARROW_ICON_CLASS, id="line-icon", style=_LINE_ICON_STYLE),
    dbc.Tooltip("Line", target="line-icon", placement="bottom"),
    html.I(className=_BOOTSTRAP_ARROW_ICON_CLASS, id="transformer-icon", style=_TRANSFORMER_ICON_STYLE),
    dbc.Tooltip("Transformer", target="transformer-icon", placement="bottom"),
    html.I(className=_BOOTSTRAP_ARROW_ICON_CLASS, id="link-icon", style=_LINK_ICON_STYLE),
    dbc.Tooltip("Link", target="link-icon", placement="bottom"),
    html.I(className=_BOOTSTRAP_ARROW_ICON_CLASS, id="generic-branch-icon", style=_GENERIC_BRANCH_ICON_STYLE),
    dbc.Tooltip("Generic Branch", target="generic-branch-icon", placement="bottom"),
    html.I(className=_BOOTSTRAP_ARROW_ICON_CLASS, id="asym-line-icon", style=_ASYM_LINE_ICON_STYLE),
    dbc.Tooltip("Asymmetrical Line", target="asym-line-icon", placement="bottom"),
    html.I(className="fas fa-ellipsis", id="open-branch-icon", style=_OPEN_BRANCH_ICON_STYLE),
    dbc.Tooltip("Open Branch", target="open-branch-icon", placement="bottom"),
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import logging
import warnings
from typing import TYPE_CHECKING

from power_grid_model_ds._core.model.arrays import (
    Branch3Array,
    BranchArray,
    LineArray,
    LinkArray,
    NodeArray,
    TransformerArray,
)

from ..arrays.base.array import FancyArray

if TYPE_CHECKING:
    from .base import Grid


logger = logging.getLogger(__name__)


def add_array_to_grid(grid: "Grid", array: FancyArray, check_max_id: bool = True) -> None:
    """See Grid.append()"""
    grid._append(array, check_max_id=check_max_id)  # noqa # pylint: disable=protected-access
    # pylint: disable=protected-access
    grid.graphs._append(array)


def add_node(grid: "Grid", node: NodeArray) -> None:
    """See Grid.add_node()"""
    warnings.warn(
        "Grid.add_node is deprecated and will be removed in a future release. Use Grid.append instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    grid._append(array=node)  # noqa # pylint: disable=protected-access
    grid.graphs.add_node_array(node_array=node)
    logging.debug(f"added node {node.id}")


def add_branch(grid: "Grid", branch: BranchArray) -> None:
    """See Grid.add_branch()"""
    warnings.warn(
        "Grid.add_branch is deprecated and will be removed in a future release. Use Grid.append instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    grid._append(array=branch)  # noqa # pylint: disable=protected-access
    grid.graphs.add_branch_array(branch_array=branch)

    logging.debug(f"added branch {branch.id} from {branch.from_node} to {branch.to_node}")


def reverse_branches(grid: "Grid", branches: BranchArray) -> None:
    """See Grid.reverse_branches()"""
    if not branches.size:
        return
    if not isinstance(branches, (LineArray, LinkArray, TransformerArray)):
        try:
            branches = grid.get_typed_branches(branches.id)
        except ValueError:
            # If the branches are not of the same type, reverse them per type (though this is slower)
            for array in grid.branch_arrays:
                grid.reverse_branches(array.filter(branches.id))
            return

    from_nodes = branches.from_node
    to_nodes = branches.to_node

    array_field = grid.find_array_field(branches.__class__)
    array = getattr(grid, array_field.name)
    array.update_by_id(branches.id, from_node=to_nodes, to_node=from_nodes)


def make_active(grid: "Grid", branch: BranchArray) -> None:
    """See Grid.make_active()"""
    array_field = grid.find_array_field(branch.__class__)
    array_attr = getattr(grid, array_field.name)
    branch_mask = array_attr.id == branch.id
    array_attr.from_status[branch_mask] = 1
    array_attr.to_status[branch_mask] = 1
    setattr(grid, array_field.name, array_attr)

    grid.graphs.make_active(branch=branch)
    logging.debug(f"activated branch {branch.id}")


def make_inactive(grid, branch: BranchArray, at_to_side: bool = True) -> None:
    """See Grid.make_inactive()"""
    array_field = grid.find_array_field(branch.__class__)
    array_attr = getattr(grid, array_field.name)
    branch_mask = array_attr.id == branch.id
    status_side = "to_status" if at_to_side else "from_status"
    array_attr[status_side][branch_mask] = 0
    setattr(grid, array_field.name, array_attr)

    grid.graphs.make_inactive(branch=branch)
    logging.debug(f"deactivated branch {branch.id}")


def delete_node(grid: "Grid", node: NodeArray) -> None:
    """See Grid.delete_node()"""
    grid.node = grid.node.exclude(id=node.id)
    grid.sym_load = grid.sym_load.exclude(node=node.id)
    grid.source = grid.source.exclude(node=node.id)

    for branch_array in grid.branch_arrays:
        matching_branches = branch_array.filter(from_node=node.id, to_node=node.id, mode_="OR")
        for branch in matching_branches:
            grid.delete_branch(branch)

    grid.graphs.delete_node(node=node)
    logging.debug(f"deleted rail {node.id}")


def delete_branch(grid: "Grid", branch: BranchArray) -> None:
    """See Grid.delete_branch()"""
    _delete_branch_array(branch=branch, grid=grid)
    grid.graphs.delete_branch(branch=branch)
    logging.debug(f"""deleted branch {branch.id.item()} from {branch.from_node.item()} to {branch.to_node.item()}""")


def delete_branch3(grid: "Grid", branch: Branch3Array) -> None:
    """See Grid.delete_branch3()"""
    _delete_branch_array(branch=branch, grid=grid)
    grid.graphs.delete_branch3(branch=branch)


def _delete_branch_array(branch: BranchArray | Branch3Array, grid: "Grid"):
    # Delete a branch or branch3 array from the grid.
    array_field = grid.find_array_field(branch.__class__)
    array_attr = getattr(grid, array_field.name)
    setattr(grid, array_field.name, array_attr.exclude(id=branch.id))

    grid.transformer_tap_regulator = grid.transformer_tap_regulator.exclude(regulated_object=branch.id)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from collections import Counter

import rustworkx as rx


def find_fundamental_cycles_rustworkx(graph):
    """Detect fundamental cycles in the graph and returns the node cycle paths.

    The nodes in cycles are found by:
    1. Creating a minimum spanning forest (MSF) of the (possibly disconnected) graph, which is a set of minimum
    spanning trees (MST's). Here, a MST is a subset of the original graph that connects all nodes with the minimum
    possible amount of edges. (Also, see https://mathworld.wolfram.com/SpanningTree.html).

    For a connected graph, all nodes are connected by edges, leading to a single MST. In that case,
    the MSF consists of a single MST. (See https://mathworld.wolfram.com/ConnectedGraph.html for more information).

    For a disconnected graph, there are multiple separated subgraphs; in other words, not all nodes are connected.
    (ALso, see https://mathworld.wolfram.com/DisconnectedGraph.html). Each of the subgraphs will then form a MST.
    The MSF is then the collection of these MST's.

    NOTE: If there are cycles in the graph, the MSF is not unique. One of the options will be chosen.
    Regardless of the chosen MSF, the algorithm works.

    2. Determining which edges in the original graph are not part of the MSF. A propery of MST's is that adding
    a single edge between nodes will always form a cycle.

    3. Looping through the unused edges and determining the path of nodes of the cycle it would form.
    This is done by determining the shortest path between the nodes on both sides of an unused edge within the MSF.
    A combination of this shortest path and the unused edge forms the cycle path. A list of these paths is returned.

    Returns:
        node_cycle_paths(list[list[[int]]): a list of node paths, which are each a list of node_ids in a path.
    """
    mst = rx.minimum_spanning_tree(graph)
    unused_edges = _find_unused_edges_rustworkx(graph, mst)
    node_cycle_paths = _get_cycle_paths_rustworkx(unused_edges, mst)

    return node_cycle_paths


def _find_unused_edges_rustworkx(full_graph, subset_graph):
    """Determine unused edges by comparing all edges with a subset of edges in the MST."""
    full_edges = Counter(full_graph.edge_list())
    subset_edges = Counter(subset_graph.edge_list())
    unused_edges = full_edges - subset_edges
    return unused_edges


def _get_cycle_paths_rustworkx(unused_edges, spanning_forest_graph):
    """Find nodes that are part of a cycle in the graph using the MST."""
    node_cycle_paths = []
    for source, target in unused_edges:
        path_mapping = rx.dijkstra_shortest_paths(spanning_forest_graph, source, target, weight_fn=lambda x: 1)

        path_nodes = list(path_mapping[target])
        path_nodes.append(source)

        node_cycle_paths.append(path_nodes)
    return node_cycle_paths



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator

from numpy._typing import NDArray

from power_grid_model_ds._core.model.arrays.pgm_arrays import Branch3Array, BranchArray, NodeArray
from power_grid_model_ds._core.model.graphs.errors import (
    GraphError,
    MissingBranchError,
    MissingNodeError,
    NoPathBetweenNodes,
)

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid


# pylint: disable=too-many-public-methods
class BaseGraphModel(ABC):
    """Base class for graph models"""

    def __init__(self, active_only=False) -> None:
        self.active_only = active_only

    @property
    @abstractmethod
    def nr_nodes(self) -> int:
        """Returns the number of nodes in the graph"""

    @property
    @abstractmethod
    def nr_branches(self) -> int:
        """Returns the number of branches in the graph"""

    @property
    def all_branches(self) -> Generator[tuple[int, int], None, None]:
        """Returns all branches in the graph."""
        return (
            (self.internal_to_external(source), self.internal_to_external(target))
            for source, target in self._all_branches()
        )

    @abstractmethod
    def external_to_internal(self, ext_node_id: int) -> int:
        """Convert external node id to internal node id (internal)

        Raises:
            MissingNodeError: if the external node id does not exist in the graph
        """

    @abstractmethod
    def internal_to_external(self, int_node_id: int) -> int:
        """Convert internal id (internal) to external node id"""

    @property
    @abstractmethod
    def external_ids(self) -> list[int]:
        """Return all external node ids

        Warning: Depending on graph engine, performance could be slow for large graphs
        """

    def has_node(self, node_id: int) -> bool:
        """Check if a node exists."""
        try:
            internal_node_id = self.external_to_internal(ext_node_id=node_id)
        except MissingNodeError:
            return False

        return self._has_node(node_id=internal_node_id)

    def in_branches(self, node_id: int) -> Generator[tuple[int, int], None, None]:
        """Return all branches that have the node as an endpoint."""
        int_node_id = self.external_to_internal(node_id)
        internal_edges = self._in_branches(int_node_id=int_node_id)
        return (
            (self.internal_to_external(source), self.internal_to_external(target)) for source, target in internal_edges
        )

    def add_node(self, ext_node_id: int, raise_on_fail: bool = True) -> None:
        """Add a node to the graph."""
        if self.has_node(ext_node_id):
            if raise_on_fail:
                raise GraphError(f"External node id '{ext_node_id}' already exists!")
            return

        self._add_node(ext_node_id)

    def delete_node(self, ext_node_id: int, raise_on_fail: bool = True) -> None:
        """Remove a node from the graph.

        Args:
            ext_node_id(int): id of the node to remove
            raise_on_fail(bool): whether to raise an error if the node does not exist. Defaults to True

        Raises:
            MissingNodeError: if the node does not exist in the graph and ``raise_on_fail=True``
        """
        try:
            internal_node_id = self.external_to_internal(ext_node_id)
        except MissingNodeError as error:
            if raise_on_fail:
                raise error
            return

        self._delete_node(node_id=internal_node_id)

    def add_node_array(self, node_array: NodeArray, raise_on_fail: bool = True) -> None:
        """Add all nodes in the node array to the graph."""
        if raise_on_fail and any(self.has_node(x) for x in node_array["id"]):
            raise GraphError("At least one node id already exists in the Graph.")
        self._add_nodes(node_array["id"].tolist())

    def delete_node_array(self, node_array: NodeArray, raise_on_fail: bool = True) -> None:
        """Delete all nodes in node_array from the graph"""
        for node in node_array:
            self.delete_node(node.id.item(), raise_on_fail=raise_on_fail)

    def has_branch(self, from_ext_node_id: int, to_ext_node_id: int) -> bool:
        """Check if a branch exists between two nodes."""
        try:
            int_from_node_id = self.external_to_internal(from_ext_node_id)
            int_to_node_id = self.external_to_internal(to_ext_node_id)
        except MissingNodeError:
            return False

        return self._has_branch(from_node_id=int_from_node_id, to_node_id=int_to_node_id)

    def add_branch(self, from_ext_node_id: int, to_ext_node_id: int) -> None:
        """Add a new branch to the graph."""
        self._add_branch(
            from_node_id=self.external_to_internal(from_ext_node_id),
            to_node_id=self.external_to_internal(to_ext_node_id),
        )

    def delete_branch(self, from_ext_node_id: int, to_ext_node_id: int, raise_on_fail: bool = True) -> None:
        """Remove an existing branch from the graph.

        Args:
            from_ext_node_id: id of the from node
            to_ext_node_id: id of the to node
            raise_on_fail: whether to raise an error if the branch does not exist

        Raises:
            MissingBranchError: if branch does not exist in the graph and ``raise_on_fail=True``
        """
        try:
            self._delete_branch(
                from_node_id=self.external_to_internal(from_ext_node_id),
                to_node_id=self.external_to_internal(to_ext_node_id),
            )
        except (MissingNodeError, MissingBranchError) as error:
            if raise_on_fail:
                raise MissingBranchError(
                    f"Branch between nodes {from_ext_node_id} and {to_ext_node_id} does NOT exist!"
                ) from error

    def add_branch_array(self, branch_array: BranchArray) -> None:
        """Add all branches in the branch array to the graph."""
        if self.active_only:
            branch_array = branch_array[branch_array.is_active]
            if not branch_array.size:
                return

        from_node_ids = self._externals_to_internals(branch_array["from_node"].tolist())
        to_node_ids = self._externals_to_internals(branch_array["to_node"].tolist())
        self._add_branches(from_node_ids, to_node_ids)

    def add_branch3_array(self, branch3_array: Branch3Array) -> None:
        """Add all branch3s in the branch3 array to the graph."""
        for branch3 in branch3_array:
            self.add_branch_array(branch3.as_branches())

    def delete_branch_array(self, branch_array: BranchArray, raise_on_fail: bool = True) -> None:
        """Delete all branches in branch_array from the graph."""
        for branch in branch_array:
            if self._branch_is_relevant(branch):
                self.delete_branch(branch.from_node.item(), branch.to_node.item(), raise_on_fail=raise_on_fail)

    def delete_branch3_array(self, branch3_array: Branch3Array, raise_on_fail: bool = True) -> None:
        """Delete all branch3s in the branch3 array from the graph."""
        for branch3 in branch3_array:
            self.delete_branch_array(branch3.as_branches(), raise_on_fail=raise_on_fail)

    @contextmanager
    def tmp_remove_nodes(self, nodes: list[int]) -> Generator:
        """Context manager that temporarily removes nodes and their branches from the graph.
        Example:
            >>> with graph.tmp_remove_nodes([1, 2, 3]):
            >>>    assert not graph.has_node(1)
            >>> assert graph.has_node(1)
        In practice, this is useful when you want to e.g. calculate the shortest path between two nodes without
        considering certain nodes.
        """
        edge_list = []
        for node in nodes:
            edge_list += list(self.in_branches(node))
            self.delete_node(node)

        yield

        for node in nodes:
            self.add_node(int(node))  # convert to int to avoid type issues when input is e.g. a numpy array
        for source, target in edge_list:
            self.add_branch(source, target)

    def get_shortest_path(self, ext_start_node_id: int, ext_end_node_id: int) -> tuple[list[int], int]:
        """Calculate the shortest path between two nodes

        Example:
            given this graph: [1] - [2] - [3] - [4]

            >>> graph.get_shortest_path(1, 4) == [1, 2, 3, 4], 3
            >>> graph.get_shortest_path(1, 1) == [1], 0

        Returns:
            tuple[list[int], int]: a tuple where the first element is a list of external nodes from start to end.
            The second element is the distance of the path in number of edges.

        Raises:
            NoPathBetweenNodes: if no path exists between the given nodes
        """
        if ext_start_node_id == ext_end_node_id:
            return [ext_start_node_id], 0

        try:
            internal_path, distance = self._get_shortest_path(
                source=self.external_to_internal(ext_start_node_id), target=self.external_to_internal(ext_end_node_id)
            )
            return self._internals_to_externals(internal_path), distance
        except NoPathBetweenNodes as e:
            raise NoPathBetweenNodes(f"No path between nodes {ext_start_node_id} and {ext_end_node_id}") from e

    def get_all_paths(self, ext_start_node_id: int, ext_end_node_id: int) -> list[list[int]]:
        """Retrieves all paths between two (external) nodes.
        Returns a list of paths, each path containing a list of external nodes.
        """
        if ext_start_node_id == ext_end_node_id:
            return []

        internal_paths = self._get_all_paths(
            source=self.external_to_internal(ext_start_node_id),
            target=self.external_to_internal(ext_end_node_id),
        )

        return [self._internals_to_externals(path) for path in internal_paths]

    def get_components(self) -> list[list[int]]:
        """Returns all separate components of the graph as lists

        If you want to get the components of the graph without certain nodes,
        use the `tmp_remove_nodes` context manager.

        Example:
        >>> with graph.tmp_remove_nodes(substation_nodes):
        >>>    components = graph.get_components()
        """
        internal_components = self._get_components()
        return [self._internals_to_externals(component) for component in internal_components]

    def get_connected(
        self, node_id: int, nodes_to_ignore: list[int] | None = None, inclusive: bool = False
    ) -> list[int]:
        """Find all nodes connected to the node_id

        Args:
            node_id: node id to start the search from
            inclusive: whether to include the given node id in the result
            nodes_to_ignore: list of node ids to ignore while traversing the graph.
                              Any nodes connected to `node_id` (solely) through these nodes will
                              not be included in the result
        Returns:
            nodes: list of node ids sorted by distance, connected to the node id
        """
        if nodes_to_ignore is None:
            nodes_to_ignore = []

        nodes = self._get_connected(
            node_id=self.external_to_internal(node_id),
            nodes_to_ignore=self._externals_to_internals(nodes_to_ignore),
            inclusive=inclusive,
        )

        return self._internals_to_externals(nodes)

    def find_first_connected(self, node_id: int, candidate_node_ids: list[int]) -> int:
        """Find the first connected node to the node_id from the candidate_node_ids

        Note:
            If multiple candidate nodes are connected to the node, the first one found is returned.
            There is no guarantee that the same candidate node will be returned each time.

        Raises:
            MissingNodeError: if no connected node is found
            ValueError: if the node_id is in candidate_node_ids
        """
        internal_node_id = self.external_to_internal(node_id)
        internal_candidates = self._externals_to_internals(candidate_node_ids)
        if internal_node_id in internal_candidates:
            raise ValueError("node_id cannot be in candidate_node_ids")
        return self.internal_to_external(self._find_first_connected(internal_node_id, internal_candidates))

    def get_downstream_nodes(self, node_id: int, start_node_ids: list[int], inclusive: bool = False) -> list[int]:
        """Find all nodes downstream of the node_id with respect to the start_node_ids

        Example:
            given this graph: [1] - [2] - [3] - [4]
            >>> graph.get_downstream_nodes(2, [1]) == [3, 4]
            >>> graph.get_downstream_nodes(2, [1], inclusive=True) == [2, 3, 4]

        args:
            node_id: node id to start the search from
            start_node_ids: list of node ids considered 'above' the node_id
            inclusive: whether to include the given node id in the result
        returns:
            list of node ids sorted by distance, downstream of to the node id
        """
        connected_node = self.find_first_connected(node_id, start_node_ids)
        path, _ = self.get_shortest_path(node_id, connected_node)
        _, upstream_node, *_ = (
            path  # path is at least 2 elements long or find_first_connected would have raised an error
        )

        return self.get_connected(node_id, [upstream_node], inclusive)

    def find_fundamental_cycles(self) -> list[list[int]]:
        """Find all fundamental cycles in the graph.
        Returns:
            list[list[int]]: list of cycles, each cycle is a list of (external) node ids
        """
        internal_cycles = self._find_fundamental_cycles()
        return [self._internals_to_externals(nodes) for nodes in internal_cycles]

    @classmethod
    def from_arrays(cls, arrays: "Grid", active_only=False) -> "BaseGraphModel":
        """Build from arrays"""
        new_graph = cls(active_only=active_only)

        new_graph.add_node_array(node_array=arrays.node, raise_on_fail=False)
        new_graph.add_branch_array(arrays.branches)
        new_graph.add_branch3_array(arrays.three_winding_transformer)

        return new_graph

    def _internals_to_externals(self, internal_nodes: list[int]) -> list[int]:
        """Convert a list of internal nodes to external nodes"""
        return [self.internal_to_external(node_id) for node_id in internal_nodes]

    def _externals_to_internals(self, external_nodes: list[int] | NDArray) -> list[int]:
        """Convert a list of external nodes to internal nodes"""
        return [self.external_to_internal(node_id) for node_id in external_nodes]

    def _branch_is_relevant(self, branch: BranchArray) -> bool:
        """Check if a branch is relevant"""
        if self.active_only:
            return branch.is_active.item()
        return True

    @abstractmethod
    def _in_branches(self, int_node_id: int) -> Generator[tuple[int, int], None, None]: ...

    @abstractmethod
    def _get_connected(self, node_id: int, nodes_to_ignore: list[int], inclusive: bool = False) -> list[int]: ...

    @abstractmethod
    def _find_first_connected(self, node_id: int, candidate_node_ids: list[int]) -> int: ...

    @abstractmethod
    def _has_branch(self, from_node_id, to_node_id) -> bool: ...

    @abstractmethod
    def _has_node(self, node_id) -> bool: ...

    @abstractmethod
    def _add_node(self, ext_node_id: int) -> None: ...

    @abstractmethod
    def _add_nodes(self, ext_node_ids: list[int]) -> None: ...

    @abstractmethod
    def _delete_node(self, node_id: int): ...

    @abstractmethod
    def _add_branch(self, from_node_id: int, to_node_id: int) -> None: ...

    @abstractmethod
    def _add_branches(self, from_node_ids: list[int], to_node_ids: list[int]) -> None: ...

    @abstractmethod
    def _delete_branch(self, from_node_id, to_node_id) -> None:
        """
        Raises:
            MissingBranchError: if the branch does not exist
        """

    @abstractmethod
    def _get_shortest_path(self, source, target): ...

    @abstractmethod
    def _get_all_paths(self, source, target) -> list[list[int]]: ...

    @abstractmethod
    def _get_components(self) -> list[list[int]]: ...

    @abstractmethod
    def _find_fundamental_cycles(self) -> list[list[int]]: ...

    @abstractmethod
    def _all_branches(self) -> Generator[tuple[int, int], None, None]: ...



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Helper functions for arrays"""

import numpy as np
from numpy.typing import ArrayLike, NDArray


def re_order(array: np.ndarray, new_order: ArrayLike, column: str = "id") -> np.ndarray:
    """Re-order an id-array by the id column so that it follows a new_order.
    Expects the new_order input to contain the same values as self.id
    """
    if column not in (array.dtype.names or ()):
        raise ValueError(f"Cannot re-order array: column {column} does not exist.")
    if not np.array_equal(np.sort(array[column]), np.sort(new_order)):
        raise ValueError(f"Cannot re-order array: mismatch between new_order and values in '{column}'-column.")

    permutation_a = np.argsort(array[column])
    permutation_b = np.argsort(new_order)
    inverse = np.empty_like(new_order, dtype=int)
    inverse[permutation_b] = np.arange(permutation_b.size)
    new_order_indices = permutation_a[inverse]
    return array[new_order_indices]


def update_by_id(array: np.ndarray, ids: ArrayLike, allow_missing: bool, **kwargs) -> NDArray[np.bool_]:
    """Update values in an array by id

    Args:
        array: the array to update
        ids: the ids to update
        allow_missing: whether to allow ids that do not exist in the array
        **kwargs: the columns to update and their new values
    Returns:
        mask: the mask on the original array for the provided ids
    """
    mask = np.isin(array["id"], ids)
    if not allow_missing:
        nr_hits = np.sum(mask)
        nr_ids = np.unique(ids).size  # ignore edge cases with duplicate ids
        if nr_hits != nr_ids:
            raise ValueError("One or more ids do not exist. Provide allow_missing=True if this is intended.")

    for name, values in kwargs.items():
        array[name][mask] = values
    return mask


def check_ids(array: np.ndarray, return_duplicates: bool = False) -> NDArray | None:
    """Check for duplicate ids within the array"""
    if "id" not in (array.dtype.names or ()):
        raise AttributeError("Array has no 'id' column.")

    unique, counts = np.unique(array["id"], return_counts=True)
    duplicate_mask = counts > 1
    duplicates = unique[duplicate_mask]

    if return_duplicates:
        return duplicates
    if duplicates.size > 0:
        raise ValueError(f"Found duplicate ids in array: {duplicates}")
    return None



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from dash import Input, Output, callback, dash_table

from power_grid_model_ds._core.visualizer.layout.selection_output import (
    SELECTION_OUTPUT_HTML,
)


@callback(
    Output("selection-output", "children"),
    Input("cytoscape-graph", "selectedNodeData"),
    Input("cytoscape-graph", "selectedEdgeData"),
)
def display_selected_element(node_data: list[dict[str, Any]], edge_data: list[dict[str, Any]]):
    """Display the tapped edge data."""
    if node_data:
        return _to_data_table(node_data.pop())
    if edge_data:
        edge_data_dict = edge_data.pop()
        del edge_data_dict["source"]  # duplicated by from_node
        del edge_data_dict["target"]  # duplicated by to_node
        del edge_data_dict["group"]  # unnecessary information
        return _to_data_table(edge_data_dict)
    return SELECTION_OUTPUT_HTML


def _to_data_table(data: dict[str, Any]):
    columns = data.keys()
    data_table = dash_table.DataTable(  # type: ignore[attr-defined]
        data=[data], columns=[{"name": key, "id": key} for key in columns], editable=False
    )
    return data_table



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
import logging
from dataclasses import Field, fields
from typing import TYPE_CHECKING

from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.utils.misc import array_equal_with_nan

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import FancyArrayContainer

logger = logging.getLogger("power_grid_model_ds.fancypy")  # public location of the container_equal function


def container_equal(
    container_a: "FancyArrayContainer",
    container_b: "FancyArrayContainer",
    ignore_extras: bool = False,
    early_exit: bool = True,
    fields_to_ignore: list[str] = None,
) -> bool:
    """
    Compares two containers for equality.

    Args:
        container_a: The first container to compare.
        container_b: The second container to compare.
        ignore_extras:
            If True,
                ignores fields present in one container_a but not in container_b.
                ignores extra columns in arrays in container_b that are not present in container_a.
        early_exit: If True, returns False on the first detected difference. False to log all differences as debug.
        fields_to_ignore: A list of field names to exclude from comparison.

    Returns:
        True if the containers are equal, False otherwise.
    """
    fields_to_ignore = fields_to_ignore or []
    is_equal = True

    for field in fields(container_a):
        if field.name in fields_to_ignore or (ignore_extras and not hasattr(container_b, field.name)):
            continue

        if not _fields_are_equal(container_a, container_b, field, ignore_extras):
            is_equal = False
            if early_exit:
                return False

    if not ignore_extras and _check_for_extra_fields(container_a, container_b):
        return False

    return is_equal


def _fields_are_equal(
    container_a: "FancyArrayContainer", container_b: "FancyArrayContainer", field: "Field", ignore_extras: bool
) -> bool:
    """Compares a single field between two containers."""
    value_a = getattr(container_a, field.name)
    value_b = getattr(container_b, field.name)
    class_name = container_a.__class__.__name__

    if isinstance(value_a, FancyArray):
        if not _check_array_equal(value_a, value_b, ignore_extras):
            logger.debug(f"Array field '{field.name}' differs between {class_name}s.")
            return False
    elif value_a != value_b:
        logger.debug(f"Field '{field.name}' differs between {class_name}s.")
        return False

    return True


def _check_for_extra_fields(container_a: "FancyArrayContainer", container_b: "FancyArrayContainer") -> bool:
    """Checks if container_b has extra fields not present in container_a."""
    fields_a = {f.name for f in fields(container_a)}
    fields_b = {f.name for f in fields(container_b)}
    extra_fields = fields_b - fields_a

    if extra_fields:
        logger.debug(f"Container {container_b.__class__.__name__} has extra fields: {extra_fields}")
        return True
    return False


def _check_array_equal(array_a: FancyArray, array_b: FancyArray, ignore_extras: bool) -> bool:
    """
    Compares two FancyArrays, optionally ignoring extra columns in array_b.
    NaN values are treated as equal.
    """
    data_a = array_a.data
    data_b = array_b.data

    if ignore_extras:
        common_columns = [col for col in array_a.columns if col in array_b.columns]
        data_a = array_a[common_columns]
        data_b = array_b[common_columns]

    return array_equal_with_nan(data_a, data_b)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Enums for Nodes"""

from enum import IntEnum


class NodeType(IntEnum):
    """Node Types
    Nodes located within a substation, are marked as Substation nodes.
    """

    UNSPECIFIED = 0
    SUBSTATION_NODE = 1



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.model.grids.base import Grid
from power_grid_model_ds._core.power_grid_model_interface import PowerGridModelInterface

__all__ = [
    "Grid",
    "GraphContainer",
    "PowerGridModelInterface",
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Serialization utilities for Grid objects using power-grid-model serialization with extensions support."""

import dataclasses
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

from power_grid_model_ds._core.model.arrays.base.array import FancyArray

if TYPE_CHECKING:
    # Import only for type checking to avoid circular imports at runtime
    from power_grid_model_ds._core.model.grids.base import Grid


G = TypeVar("G", bound="Grid")


logger = logging.getLogger(__name__)


def serialize_to_json(grid: G, path: Path, strict: bool = True, **kwargs) -> Path:
    """Save a Grid object to JSON format using power-grid-model serialization with extensions support.

    Args:
        grid: The Grid object to serialize
        path: The file path to save to
        strict: Whether to raise an error if the grid object is not serializable.
        **kwargs: Keyword arguments forwarded to json.dump (for example, indent, sort_keys,
            ensure_ascii, etc.).
    Returns:
        Path: The path where the file was saved
    """
    path.parent.mkdir(parents=True, exist_ok=True)

    serialized_data = {}

    for field in dataclasses.fields(grid):
        if field.name in ["graphs"]:
            continue

        field_value = getattr(grid, field.name)

        if isinstance(field_value, FancyArray):
            serialized_data[field.name] = _serialize_array(field_value)
            continue

        if _is_serializable(field_value, strict):
            serialized_data[field.name] = field_value

    # Store in a wrapper for PGM compatibility
    json_data = {"data": serialized_data}

    with open(path, "w", encoding="utf-8") as f:
        json.dump(json_data, f, **kwargs)

    return path


def deserialize_from_json(path: Path, target_grid_class: type[G]) -> G:
    """Load a Grid object from JSON format with cross-type loading support.

    Args:
        path: The file path to load from
        target_grid_class: Grid class to load into.

    Returns:
        Grid: The deserialized Grid object of the specified target class
    """
    with open(path, "r", encoding="utf-8") as f:
        json_data = json.load(f)

    grid = target_grid_class.empty()
    _restore_grid_values(grid, json_data["data"])
    graph_class = grid.graphs.__class__
    grid.graphs = graph_class.from_arrays(grid)
    return grid


def _restore_grid_values(grid: G, json_data: dict) -> None:
    """Restore arrays to the grid."""
    for attr_name, attr_values in json_data.items():
        if not hasattr(grid, attr_name):
            logger.warning(f"Unexpected attribute '{attr_name}'")
            continue

        grid_attr = getattr(grid, attr_name)
        attr_class = grid_attr.__class__
        if isinstance(grid_attr, FancyArray):
            array = _deserialize_array(array_data=attr_values, array_class=attr_class)
            setattr(grid, attr_name, array)
            continue

        # load other values
        setattr(grid, attr_name, attr_class(attr_values))


def _serialize_array(array: FancyArray) -> list[dict[str, Any]]:
    return [{name: record[name].item() for name in array.columns} for record in array]


def _deserialize_array(array_data: list[dict[str, Any]], array_class: type[FancyArray]) -> FancyArray:
    if not array_data:
        return array_class()

    array_columns = set(array_class.get_dtype().names)

    data_as_dict_of_lists: dict[str, Any] = {}
    for column in array_columns:
        column_data = [row[column] for row in array_data if column in row]
        if len(column_data) not in [0, len(array_data)]:
            raise ValueError(
                f"Some records in column '{column}' have missing values. "
                f"For defaulted columns, either provide all values or none."
            )
        if column_data:
            data_as_dict_of_lists[column] = column_data

    all_columns_in_array_data = set().union(*(row.keys() for row in array_data))
    extra_columns = all_columns_in_array_data - array_columns
    if extra_columns:
        logger.warning(f"Ignoring extra columns {extra_columns} from array data for {array_class.__name__}.")
    return array_class(**data_as_dict_of_lists)


def _is_serializable(value: Any, strict: bool) -> bool:
    # Check if a value is JSON serializable.
    try:
        json.dumps(value)
    except TypeError as error:
        msg = f"Failed to serialize '{value}'. You can set strict=False to ignore this attribute."
        if strict:
            raise TypeError(msg) from error
        logger.warning(msg)
        return False
    return True



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Appliance data types"""

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.dtypes.id import Id


class Appliance(Id):
    """Appliance data type"""

    node: NDArray[np.int32]  # id of the coupled node
    status: NDArray[np.int8]  # connection status to the node


class Source(Appliance):
    """Source data type"""

    u_ref: NDArray[np.float64]  # reference voltage


class SymLoad(Appliance):
    """SymLoad data type"""

    type: NDArray[np.int8]  # load type
    p_specified: NDArray[np.float64]  # specified active power
    q_specified: NDArray[np.float64]  # specified reactive power


class SymGen(Appliance):
    """SymGen data type"""

    type: NDArray[np.int_]  # load type
    p_specified: NDArray[np.float64]  # specified active power
    q_specified: NDArray[np.float64]  # specified reactive power



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import html

from power_grid_model_ds._core.visualizer.layout.header_config import CONFIG_ELEMENTS
from power_grid_model_ds._core.visualizer.layout.header_legenda import LEGENDA_ELEMENTS, LEGENDA_STYLE
from power_grid_model_ds._core.visualizer.layout.header_search import SEARCH_ELEMENTS

_MENU_BUTTON_STYLE_CLASS = "me-2 btn-outline-primary"

_LEFT_COLUMN_HTML = dbc.Col(
    [
        dbc.Button("Legend", id="btn-legend", className=_MENU_BUTTON_STYLE_CLASS),
        dbc.Button("Search", id="btn-search", className=_MENU_BUTTON_STYLE_CLASS),
        dbc.Button("Config", id="btn-config", className=_MENU_BUTTON_STYLE_CLASS),
    ],
    id="header-left-col",
    width=5,
    style={
        "display": "flex",
        "align-items": "center",
        "justify-content": "center",
        "border-right": "1px solid white",
    },
)


_RIGHT_COLUMN_STYLE = {
    "display": "flex",
    "align-items": "center",
    "width": "100%",
}


CONFIG_DIV = html.Div(CONFIG_ELEMENTS, style=_RIGHT_COLUMN_STYLE | {"justify-content": "space-between"})
SEARCH_DIV = html.Div(SEARCH_ELEMENTS, style=_RIGHT_COLUMN_STYLE | {"justify-content": "center"})
LEGENDA_DIV = html.Div(LEGENDA_ELEMENTS, style=_RIGHT_COLUMN_STYLE | LEGENDA_STYLE)

_RIGHT_COLUMN_HTML = dbc.Col(
    [LEGENDA_DIV, SEARCH_DIV, CONFIG_DIV],
    id="header-right-col",
    width=7,
)

HEADER_HTML = dbc.Row(
    [
        _LEFT_COLUMN_HTML,
        _RIGHT_COLUMN_HTML,
    ],
    style={
        "background-color": "#343a40",
        "color": "white",
        "padding": "1rem 0",
        "margin": 0,
        "height": "90px",
        "width": "100%",
    },
    align="center",
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
import logging
from dataclasses import fields
from typing import TYPE_CHECKING, Type, TypeVar

from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel
from power_grid_model_ds._core.model.graphs.models.rustworkx import RustworkxGraphModel

if TYPE_CHECKING:
    from .base import Grid

G = TypeVar("G", bound="Grid")


logger = logging.getLogger(__name__)


def create_grid_from_extended_grid(grid_class: type[G], extended: G) -> G:
    """See Grid.from_extended()"""
    new_grid = grid_class.empty()

    # Add nodes first, so that branches can reference them
    new_grid.append(new_grid.node.__class__.from_extended(extended.node))

    for field in fields(new_grid):
        if field.name == "node":
            continue  # already added
        if isinstance(field.type, type) and issubclass(field.type, FancyArray):
            extended_array = getattr(extended, field.name)
            new_array = field.type.from_extended(extended_array)
            new_grid.append(new_array, check_max_id=False)
    return new_grid


def create_empty_grid(grid_class: Type[G], graph_model: type[BaseGraphModel] = RustworkxGraphModel) -> G:
    """See Grid.empty()"""
    empty_fields = grid_class._get_empty_fields()  # noqa # pylint: disable=protected-access
    empty_fields["graphs"] = GraphContainer.empty(graph_model=graph_model)
    return grid_class(**empty_fields)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dataclasses
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from power_grid_model_ds._core import fancypy as fp
from power_grid_model_ds._core.model.arrays import BranchArray
from power_grid_model_ds._core.model.arrays.base.errors import RecordDoesNotExist
from power_grid_model_ds._core.model.enums.nodes import NodeType

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid


def get_branches(grid: "Grid") -> BranchArray:
    """see Grid.get_branches()"""
    branch_dtype = BranchArray.get_dtype()
    branches = BranchArray()
    for array in grid.branch_arrays:
        new_branch = BranchArray(data=array.data[list(branch_dtype.names)])
        branches = fp.concatenate(branches, new_branch)
    return branches


def get_branch_arrays(grid: "Grid") -> list[BranchArray]:
    """see Grid.get_branch_arrays()"""
    branch_arrays = []
    for field in dataclasses.fields(grid):
        array = getattr(grid, field.name)
        if isinstance(array, BranchArray):
            branch_arrays.append(array)
    return branch_arrays


def get_typed_branches(grid: "Grid", branch_ids: list[int] | npt.NDArray[np.int32]) -> BranchArray:
    """see Grid.get_typed_branches()"""
    if not np.any(branch_ids):
        raise ValueError("No branch_ids provided.")
    for branch_array in grid.branch_arrays:
        array = branch_array.filter(branch_ids)
        if 0 < array.size != len(branch_ids):
            raise ValueError("Branches are not of the same type.")
        if array.size:
            return array
    raise RecordDoesNotExist(f"Branches {branch_ids} not found in grid.")


def get_nearest_substation_node(grid: "Grid", node_id: int):
    """See Grid.get_nearest_substation_node()"""
    connected_nodes = grid.graphs.active_graph.get_connected(node_id=node_id, inclusive=True)
    substation_nodes = grid.node.filter(node_type=NodeType.SUBSTATION_NODE.value)

    for node in connected_nodes:
        if node in substation_nodes.id:
            return substation_nodes.get(node)
    raise RecordDoesNotExist(f"No {NodeType.SUBSTATION_NODE.name} connected to node {node_id}")


def get_downstream_nodes(grid: "Grid", node_id: int, inclusive: bool = False):
    """See Grid.get_downstream_nodes()"""
    substation_nodes = grid.node.filter(node_type=NodeType.SUBSTATION_NODE.value)

    if node_id in substation_nodes.id:
        raise NotImplementedError("get_downstream_nodes is not implemented for substation nodes!")

    return grid.graphs.active_graph.get_downstream_nodes(
        node_id=node_id, start_node_ids=list(substation_nodes.id), inclusive=inclusive
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0


class GraphError(Exception):
    """Raised when there is an error in the graph"""


class MissingNodeError(GraphError):
    """Raised when a node is missing"""


class MissingBranchError(GraphError):
    """Raised when a branch is missing"""


class NoPathBetweenNodes(GraphError):
    """Raised when there is no path between two nodes"""



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
from enum import Enum

import dash_bootstrap_components as dbc
from dash import dcc, html

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS

NODE_SCALE_HTML = [
    html.I(className="fas fa-circle", style={"color": CYTO_COLORS["node"], "margin-right": "10px"}),
    dcc.Input(
        id="node-scale-input",
        type="number",
        value=1,
        min=0.1,
        step=0.1,
        style={"width": "75px"},
    ),
    html.Span(style={"margin-right": "10px"}),
]

EDGE_SCALE_HTML = [
    html.I(className="fas fa-arrow-right-long", style={"color": CYTO_COLORS["line"], "margin-right": "10px"}),
    dcc.Input(
        id="edge-scale-input",
        type="number",
        value=1,
        min=0.1,
        step=0.1,
        style={"width": "75px"},
    ),
]

_SCALING_DIV = html.Div(NODE_SCALE_HTML + EDGE_SCALE_HTML, style={"margin": "0 20px 0 10px"})


class LayoutOptions(Enum):
    """Cytoscape layout options."""

    RANDOM = "random"
    CIRCLE = "circle"
    CONCENTRIC = "concentric"
    GRID = "grid"
    COSE = "cose"
    BREADTHFIRST = "breadthfirst"


_LAYOUT_DROPDOWN = html.Div(
    dcc.Dropdown(
        id="dropdown-update-layout",
        placeholder="Select layout",
        value=LayoutOptions.BREADTHFIRST.value,
        clearable=False,
        options=[{"label": option.value, "value": option.value} for option in LayoutOptions],  # type: ignore[arg-type]
        style={"width": "200px"},
    ),
    style={"margin": "0 20px 0 10px", "color": "black"},
)


_ARROWS_CHECKBOX = dbc.Checkbox(
    id="show-arrows",
    label="Show arrows",
    value=True,
    label_style={"color": "white"},
    style={"margin-top": "10px"},
)

CONFIG_ELEMENTS = [_LAYOUT_DROPDOWN, _ARROWS_CHECKBOX, _SCALING_DIV]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0


from dash import Input, Output, State, callback
from dash.exceptions import PreventUpdate

from power_grid_model_ds._core.visualizer.layout.cytoscape_styling import BRANCH_WIDTH, NODE_SIZE
from power_grid_model_ds._core.visualizer.typing import STYLESHEET


@callback(
    Output("stylesheet-store", "data", allow_duplicate=True),
    Output("cytoscape-graph", "stylesheet", allow_duplicate=True),
    Input("node-scale-input", "value"),
    Input("edge-scale-input", "value"),
    State("stylesheet-store", "data"),
    prevent_initial_call=True,
)
def scale_elements(node_scale: float, edge_scale: float, stylesheet: STYLESHEET) -> tuple[STYLESHEET, STYLESHEET]:
    """Callback to scale the elements of the graph."""
    if stylesheet is None:
        raise PreventUpdate
    if node_scale == 1 and edge_scale == 1:
        raise PreventUpdate
    new_stylesheet = stylesheet.copy()
    edge_style = {
        "selector": "edge",
        "style": {
            "width": BRANCH_WIDTH * edge_scale,
        },
    }
    new_stylesheet.append(edge_style)
    node_style = {
        "selector": "node",
        "style": {
            "height": NODE_SIZE * node_scale,
            "width": NODE_SIZE * node_scale,
        },
    }
    new_stylesheet.append(node_style)

    return new_stylesheet, new_stylesheet


@callback(Output("cytoscape-graph", "layout"), Input("dropdown-update-layout", "value"), prevent_initial_call=True)
def update_layout(layout):
    """Callback to update the layout of the graph."""
    return {"name": layout, "animate": True}


@callback(
    Output("cytoscape-graph", "stylesheet", allow_duplicate=True),
    Input("show-arrows", "value"),
    State("cytoscape-graph", "stylesheet"),
    prevent_initial_call=True,
)
def update_arrows(show_arrows, current_stylesheet):
    """Callback to update the arrow style of edges in the graph."""
    selectors = [rule["selector"] for rule in current_stylesheet]
    index = selectors.index("edge")
    edge_style = current_stylesheet[index]["style"]

    edge_style["target-arrow-shape"] = "triangle" if show_arrows else "none"
    return current_stylesheet



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""A collection of errors"""


class ArrayDefinitionError(Exception):
    """Raised when an array is defined incorrectly"""


class RecordDoesNotExist(IndexError):
    """Raised when a record is accessed that does not exist"""


class MultipleRecordsReturned(IndexError):
    """Raised when unexpectedly, multiple records are found."""



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from dash import dcc, html

SELECTION_OUTPUT_HEADER_STYLE = {"margin": "20px 0 10px 0"}
_SELECTION_OUTPUT_STYLE = {"overflowX": "scroll", "textAlign": "center", "margin": "10px"}

SELECTION_OUTPUT_HTML = html.Div(
    dcc.Markdown("Click on a **node** or **edge** to display its attributes.", style=SELECTION_OUTPUT_HEADER_STYLE),
    id="selection-output",
    style=_SELECTION_OUTPUT_STYLE,
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Base grid classes"""

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self, Type, TypeVar

import numpy as np
import numpy.typing as npt

from power_grid_model_ds._core.model.arrays import (
    AsymCurrentSensorArray,
    AsymLineArray,
    AsymPowerSensorArray,
    AsymVoltageSensorArray,
    Branch3Array,
    BranchArray,
    GenericBranchArray,
    LineArray,
    LinkArray,
    NodeArray,
    SourceArray,
    SymCurrentSensorArray,
    SymGenArray,
    SymLoadArray,
    SymPowerSensorArray,
    SymVoltageSensorArray,
    ThreeWindingTransformerArray,
    TransformerArray,
    TransformerTapRegulatorArray,
)
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.containers.base import FancyArrayContainer
from power_grid_model_ds._core.model.containers.helpers import container_equal
from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.model.graphs.models import RustworkxGraphModel
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel
from power_grid_model_ds._core.model.grids._feeders import set_feeder_ids
from power_grid_model_ds._core.model.grids._helpers import create_empty_grid, create_grid_from_extended_grid
from power_grid_model_ds._core.model.grids._modify import (
    add_array_to_grid,
    add_branch,
    add_node,
    delete_branch,
    delete_branch3,
    delete_node,
    make_active,
    make_inactive,
    reverse_branches,
)
from power_grid_model_ds._core.model.grids._search import (
    get_branch_arrays,
    get_branches,
    get_downstream_nodes,
    get_nearest_substation_node,
    get_typed_branches,
)
from power_grid_model_ds._core.model.grids.serialization.json import deserialize_from_json, serialize_to_json
from power_grid_model_ds._core.model.grids.serialization.pickle import load_grid_from_pickle, save_grid_to_pickle
from power_grid_model_ds._core.model.grids.serialization.string import (
    deserialize_from_str,
    deserialize_from_txt_file,
    serialize_to_str,
)

G = TypeVar("G", bound="Grid")

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods


@dataclass
class Grid(FancyArrayContainer):
    """Grid object containing the entire network and interface to interact with it.

    Examples:

        >>> from power_grid_model_ds import Grid
        >>> grid = Grid.empty()
        >>> grid
    """

    graphs: GraphContainer
    """The graph representations of the grid."""

    # nodes
    node: NodeArray

    # branches
    transformer: TransformerArray
    three_winding_transformer: ThreeWindingTransformerArray
    line: LineArray
    link: LinkArray
    generic_branch: GenericBranchArray
    asym_line: AsymLineArray

    source: SourceArray
    sym_load: SymLoadArray
    sym_gen: SymGenArray

    # regulators
    transformer_tap_regulator: TransformerTapRegulatorArray

    # sensors
    sym_power_sensor: SymPowerSensorArray
    sym_voltage_sensor: SymVoltageSensorArray
    sym_current_sensor: SymCurrentSensorArray
    asym_power_sensor: AsymPowerSensorArray
    asym_voltage_sensor: AsymVoltageSensorArray
    asym_current_sensor: AsymCurrentSensorArray

    def __str__(self) -> str:
        """Serialize grid to a string.
        Compatible with https://csacademy.com/app/graph_editor/
        """
        return serialize_to_str(self)

    def __eq__(self, other: Any) -> bool:
        """Check if two grids are equal.

        Note: differences in graphs are ignored in this comparison.
        """
        if not isinstance(other, self.__class__):
            return False
        return container_equal(self, other, ignore_extras=False, early_exit=True, fields_to_ignore=["graphs"])

    @classmethod
    def empty(cls: Type[G], graph_model: type[BaseGraphModel] = RustworkxGraphModel) -> G:
        """Create an empty grid

        Args:
            graph_model (type[BaseGraphModel], optional): The graph model to use. Defaults to RustworkxGraphModel.

        Returns:
            Grid: An empty grid
        """
        return create_empty_grid(cls, graph_model=graph_model)

    @classmethod
    # pylint: disable=arguments-differ
    def from_cache(cls: Type[Self], cache_path: Path, load_graphs: bool = True) -> Self:
        """Read from cache and build .graphs from arrays

        Args:
            cache_path (Path): The path to the cache
            load_graphs (bool, optional): Whether to load the graphs. Defaults to True.

        Returns:
            G: The grid loaded from cache
        """
        warnings.warn(
            "Grid.from_cache() is deprecated and will be removed in a future version. Use deserialize() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return load_grid_from_pickle(cls, cache_path=cache_path, load_graphs=load_graphs)

    @classmethod
    def from_txt(cls: Type[G], *args: str) -> G:
        """Build a grid from a list of strings

        See the documentation for the expected format of the txt_lines

        Args:
            *args (str): The lines of the grid

        Examples:
            >>> Grid.from_txt("1 2", "2 3", "3 4 transformer", "4 5", "S1 6")
            alternative: Grid.from_txt("1 2\n2 3\n3 4 transformer\n4 5\nS1 6")
        """
        return deserialize_from_str(cls, *args)

    @classmethod
    # pylint: disable=arguments-differ
    def from_txt_file(cls: Type[G], txt_file_path: Path) -> G:
        """Load grid from txt file

        Args:
            txt_file_path (Path): The path to the txt file
        """
        return deserialize_from_txt_file(cls, txt_file_path)

    @classmethod
    def from_extended(cls: Type[G], extended: G) -> G:
        """Create a grid from an extended Grid object."""
        return create_grid_from_extended_grid(cls, extended=extended)

    @property
    def branches(self) -> BranchArray:
        """Converts all branch arrays into a single BranchArray."""
        return get_branches(self)

    @property
    def branch_arrays(self) -> list[BranchArray]:
        """Returns all branch arrays"""
        return get_branch_arrays(self)

    def append(self, array: FancyArray, check_max_id: bool = True):
        """Append an array to the grid. Both 'grid arrays' and 'grid.graphs' will be updated.

        Args:
            array (FancyArray): The array to append.
            check_max_id (bool, optional): Whether to check if the array id is the maximum id. Defaults to True.
        """
        return add_array_to_grid(self, array=array, check_max_id=check_max_id)

    def add_branch(self, branch: BranchArray) -> None:
        """Add a branch to the grid

        Args:
            branch (BranchArray): The branch to add
        """
        return add_branch(self, branch)

    def delete_branch(self, branch: BranchArray) -> None:
        """Remove a branch from the grid

        Args:
            branch (BranchArray): The branch to remove
        """
        return delete_branch(self, branch=branch)

    def delete_branch3(self, branch: Branch3Array) -> None:
        """Remove a branch3 from the grid

        Args:
            branch (Branch3Array): The branch3 to remove
        """
        return delete_branch3(self, branch=branch)

    def add_node(self, node: NodeArray) -> None:
        """Add a new node to the grid

        Args:
            node (NodeArray): The node to add
        """
        return add_node(self, node=node)

    def delete_node(self, node: NodeArray) -> None:
        """Remove a node from the grid

        Args:
            node (NodeArray): The node to remove
        """
        return delete_node(self, node=node)

    def make_active(self, branch: BranchArray) -> None:
        """Make a branch active

        Args:
            branch (BranchArray): The branch to make active
        """
        return make_active(self, branch=branch)

    def make_inactive(self, branch: BranchArray, at_to_side: bool = True) -> None:
        """Make a branch inactive. This is done by setting from or to status to 0.

        Args:
            branch (BranchArray): The branch to make inactive
            at_to_side (bool, optional): Whether to deactivate the to_status instead of the from_status.
            Defaults to True.
        """
        return make_inactive(self, branch=branch, at_to_side=at_to_side)

    def set_feeder_ids(self):
        """Determines all feeder groups in the network and gives them a feeder identifier

        feeder_branch_id := All assets being connected through the same feeding branch(es) on the substation
        feeder_node_id := All assets connected to the same feeder node (substation)

        Example:
        Nodes Topology
        (SUBSTATION) 101 --- 102 --- 103 -|- 104 --- 105 --- 101 (SUBSTATION)
                             {-}
                             106

        Branches Topology:
        (SUBSTATION) *** 201 *** 202 *** 203 *** 601 *** 204 *** (SUBSTATION)
                             301
                             ***

             Substation ID | Feeder ID
        101 | -1           | -1
        102 | 101          | 201
        103 | 101          | 201
        104 | 101          | 204
        105 | 101          | 204
        106 | 101          | 201
        201 | 101          | 201
        202 | 101          | 201
        203 | 101          | -1
        204 | 101          | 204
        301 | 101          | 201
        601 | 101          | 204
        """
        return set_feeder_ids(grid=self)

    def get_typed_branches(self, branch_ids: list[int] | npt.NDArray[np.int32]) -> BranchArray:
        """Find a matching LineArray, LinkArray or TransformerArray for the given branch_ids

        Raises:
            ValueError:
                - If no branch_ids are provided.
                - If not all branch_ids are of the same type.
        """
        return get_typed_branches(self, branch_ids)

    def reverse_branches(self, branches: BranchArray):
        """Reverse the direction of the branches."""
        return reverse_branches(self, branches)

    def get_branches_in_path(self, nodes_in_path: list[int]) -> BranchArray:
        """Returns all branches within a path of nodes

        Args:
            nodes_in_path (list[int]): The nodes in the path

        Returns:
            BranchArray: The branches in the path
        """
        return self.branches.filter(from_node=nodes_in_path, to_node=nodes_in_path, from_status=1, to_status=1)

    def get_nearest_substation_node(self, node_id: int):
        """Find the nearest substation node.

        Args:
            node_id(int): The id of the node to find the nearest substation node for.

        Returns:
            NodeArray: The nearest substation node.

        Raises:
            RecordDoesNotExist: If no substation node is connected to the input node.
        """
        return get_nearest_substation_node(self, node_id=node_id)

    def get_downstream_nodes(self, node_id: int, inclusive: bool = False):
        """Get the downstream nodes from a node.
        Assuming each node has a single feeding substation and the grid is radial

        Example:
            given this graph: [1] - [2] - [3] - [4], with 1 being a substation node

            >>> graph.get_downstream_nodes(2) == [3, 4]
            >>> graph.get_downstream_nodes(3) == [4]
            >>> graph.get_downstream_nodes(3, inclusive=True) == [3, 4]

        Args:
            node_id(int): The id of the node to get the downstream nodes from.
            inclusive(bool): Whether to include the input node in the result.

        Raises:
            NotImplementedError: If the input node is a substation node.

        Returns:
            list[int]: The downstream nodes.
        """
        return get_downstream_nodes(self, node_id=node_id, inclusive=inclusive)

    def cache(self, cache_dir: Path, cache_name: str, compress: bool = True):
        """Cache Grid to a folder using pickle format.

        Note: Consider using serialize() for better
        interoperability and standardized format.

        Args:
            cache_dir (Path): The directory to save the cache to.
            cache_name (str): The name of the cache.
            compress (bool, optional): Whether to compress the cache. Defaults to True.
        """
        warnings.warn(
            "grid.cache() is deprecated and will be removed in a future version. Use grid.serialize() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return save_grid_to_pickle(self, cache_dir=cache_dir, cache_name=cache_name, compress=compress)

    def serialize(self, path: Path, **kwargs) -> Path:
        """Serialize the grid.

        Args:
            path: Destination file path to write JSON to.
            **kwargs: Additional keyword arguments forwarded to ``json.dump``
        Returns:
            Path: The path where the file was saved.
        """
        return serialize_to_json(grid=self, path=path, strict=True, **kwargs)

    @classmethod
    def deserialize(cls: Type[Self], path: Path) -> Self:
        """Deserialize the grid."""
        return deserialize_from_json(path=path, target_grid_class=cls)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import html

SPAN_TEXT_STYLE = {"color": "white", "margin-right": "8px", "font-weight": "bold", "text-shadow": "0 0 5px #000"}
_INPUT_STYLE = {"width": "150px", "display": "inline-block"}
# Create your form components
GROUP_INPUT = dbc.Select(
    id="search-form-group-input",
    options=[
        {"label": "node", "value": "node"},
        {"label": "line", "value": "line"},
        {"label": "link", "value": "link"},
        {"label": "transformer", "value": "transformer"},
        {"label": "branch", "value": "branch"},
    ],
    value="node",  # Default value
    style=_INPUT_STYLE,
)

COLUMN_INPUT = dbc.Select(
    id="search-form-column-input",
    options=[{"label": "id", "value": "id"}],
    value="id",  # Default value
    style=_INPUT_STYLE,
)

VALUE_INPUT = dbc.Input(id="search-form-value-input", placeholder="Enter value", type="text", style=_INPUT_STYLE)

OPERATOR_INPUT = dbc.Select(
    id="search-form-operator-input",
    options=[
        {"label": "=", "value": "="},
        {"label": "<", "value": "<"},
        {"label": ">", "value": ">"},
        {"label": "!=", "value": "!="},
    ],
    value="=",  # Default value
    style={"width": "60px", "display": "inline-block", "margin": "0 8px"},
)


# Arrange as a sentence
SEARCH_ELEMENTS = [
    html.Div(
        [
            html.Span("Search ", style=SPAN_TEXT_STYLE),
            GROUP_INPUT,
            html.Span(" with ", style=SPAN_TEXT_STYLE),
            COLUMN_INPUT,
            OPERATOR_INPUT,
            VALUE_INPUT,
        ]
    )
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import itertools
from pathlib import Path
from typing import TYPE_CHECKING, Generic, Type, TypeVar

from power_grid_model import ComponentType

from power_grid_model_ds._core.model.enums.nodes import NodeType

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid

G = TypeVar("G", bound="Grid")


def serialize_to_str(grid: "Grid") -> str:
    """See Grid.__str__()"""
    grid_str = ""

    for transformer3 in grid.three_winding_transformer:
        nodes = [transformer3.node_1.item(), transformer3.node_2.item(), transformer3.node_3.item()]
        for combo in itertools.combinations(nodes, 2):
            grid_str += f"S{combo[0]} S{combo[1]} {transformer3.id.item()},3-transformer\n"

    for branch in grid.branches:
        from_node = grid.node.get(id=branch.from_node).record
        to_node = grid.node.get(id=branch.to_node).record

        from_node_str = f"S{from_node.id}" if from_node.node_type == NodeType.SUBSTATION_NODE else str(from_node.id)
        to_node_str = f"S{to_node.id}" if to_node.node_type == NodeType.SUBSTATION_NODE else str(to_node.id)

        suffix_str = str(branch.id.item())
        if branch.from_status.item() == 0 or branch.to_status.item() == 0:
            suffix_str = f"{suffix_str},open"

        if branch.id in grid.transformer.id:
            suffix_str = f"{suffix_str},{ComponentType.transformer.value}"
        elif branch.id in grid.link.id:
            suffix_str = f"{suffix_str},{ComponentType.link.value}"
        elif branch.id in grid.line.id:
            pass  # no suffix needed
        elif branch.id in grid.generic_branch.id:
            suffix_str = f"{suffix_str},{ComponentType.generic_branch.value}"
        elif branch.id in grid.asym_line.id:
            suffix_str = f"{suffix_str},{ComponentType.asym_line.value}"
        else:
            raise ValueError(f"Branch {branch.id} is not a transformer, link, line, generic_branch or asym_line")

        grid_str += f"{from_node_str} {to_node_str} {suffix_str}\n"
    return grid_str


def deserialize_from_str(grid_class: type[G], *args: str) -> G:
    """See Grid.from_txt()"""
    return _TextSource(grid_class).load_from_txt(*args)


def deserialize_from_txt_file(grid_class: type[G], txt_file_path: Path) -> G:
    """See Grid.from_txt_file()"""
    with open(txt_file_path, "r", encoding="utf-8") as f:
        txt_lines = f.readlines()
    return deserialize_from_str(grid_class, *txt_lines)


class _TextSource(Generic[G]):
    """Class for handling text sources.

    Text sources are only intended for test purposes so that a grid can quickly be designed from a text file.
    Moreover, these text sources are compatible with the grid editor at https://csacademy.com/app/graph_editor/

    Example of a text file:
        S1 2
        2 3
        3 4 transformer
        4 5
        S1 7

    See docs/examples/3_drawing_a_grid.md for more information.
    """

    def __init__(self, grid_class: Type[G]):
        self.grid: G = grid_class.empty()

    def load_from_txt(self, *args: str) -> G:
        """Load a grid from text"""

        text_lines = [line for arg in args for line in arg.strip().split("\n")]

        txt_nodes, txt_branches = self.read_txt(text_lines)
        self.add_nodes(txt_nodes)
        self.add_branches(txt_branches)
        self.grid.set_feeder_ids()
        return self.grid

    @staticmethod
    def read_txt(txt_lines: list[str]) -> tuple[set, dict]:
        """Extract assets from text"""

        txt_nodes = set()
        txt_branches = {}
        for text_line in txt_lines:
            if not text_line.strip() or text_line.startswith("#"):
                continue  # skip empty lines and comments
            try:
                from_node_str, to_node_str, *comments = text_line.strip().split()
            except ValueError as err:
                raise ValueError(f"Text line '{text_line}' is invalid. Skipping...") from err
            comments = comments[0].split(",") if comments else []

            txt_nodes |= {from_node_str, to_node_str}
            txt_branches[(from_node_str, to_node_str)] = comments
        return txt_nodes, txt_branches

    def add_nodes(self, nodes: set[str]):
        """Add nodes to the grid"""
        source_nodes = {int(node[1:]) for node in nodes if node.startswith("S")}
        regular_nodes = {int(node) for node in nodes if not node.startswith("S")}

        if source_nodes.intersection(regular_nodes):
            raise ValueError("Source nodes and regular nodes have overlapping ids")

        for node_id in source_nodes:
            new_node = self.grid.node.empty(1)
            new_node.id = node_id
            new_node.node_type = NodeType.SUBSTATION_NODE
            self.grid.append(new_node, check_max_id=False)

        for node_id in regular_nodes:
            new_node = self.grid.node.empty(1)
            new_node.id = node_id
            self.grid.append(new_node, check_max_id=False)

    def add_branches(self, branches: dict[tuple[str, str], list[str]]):
        """Add branches to the grid"""
        for branch, comments in branches.items():
            self.add_branch(branch, comments)

    def add_branch(self, branch: tuple[str, str], comments: list[str]):
        """Add a branch to the grid"""
        from_node_str, to_node_str = branch
        from_node = int(from_node_str.replace("S", ""))
        to_node = int(to_node_str.replace("S", ""))

        if ComponentType.transformer.value in comments:
            new_branch = self.grid.transformer.empty(1)
        elif ComponentType.link.value in comments:
            new_branch = self.grid.link.empty(1)
        elif ComponentType.generic_branch.value in comments:
            new_branch = self.grid.generic_branch.empty(1)
        elif ComponentType.asym_line.value in comments:
            new_branch = self.grid.asym_line.empty(1)
        else:  # assume it is a line
            new_branch = self.grid.line.empty(1)

        branch_ids = [branch_id for branch_id in comments if branch_id.isdigit()]
        if branch_ids:
            if len(branch_ids) > 1:
                raise ValueError(f"Multiple branch ids found in row {branch} {','.join(comments)}")
            new_branch.id = int(branch_ids[0])

        new_branch.from_node = from_node
        new_branch.to_node = to_node
        new_branch.from_status = 1
        if "open" in comments:
            new_branch.to_status = 0
        else:
            new_branch.to_status = 1
        self.grid.append(new_branch, check_max_id=False)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Module for array to string conversion."""

from typing import TYPE_CHECKING, Optional

import numpy as np

from power_grid_model_ds._core import fancypy as fp

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.arrays.base.array import FancyArray


def convert_array_to_string(array: "FancyArray", rows: int = 10, column_width: int | str = "auto") -> str:
    """Return a string representation of the array as a table.

    Args:
        column_width: the width of each column in the table. Will be determined by its contents if set to "auto".
        rows: the number of rows to show. If the array is larger than this, the middle rows are hidden.
    """
    start_rows, end_rows = _get_start_and_end_rows(array, rows)
    if end_rows is not None:
        rows_to_print = fp.concatenate(start_rows, end_rows)
    else:
        rows_to_print = start_rows

    match column_width:
        case "auto":
            column_widths = _determine_column_widths(rows_to_print)
        case int():
            column_widths = [(column, column_width) for column in array.dtype.names]
        case _:
            raise NotImplementedError(f"column_width={column_width} is not supported. Use 'auto' or int.")

    header = "|".join(f"{_center_and_truncate(column, width)}" for column, width in column_widths) + "\n"
    if end_rows is None:
        body = _rows_to_strings(rows_to_print, column_widths)
        return header + "\n".join(body)

    start_rows = _rows_to_strings(start_rows, column_widths)
    cutoff_line = f"(..{array.size - rows_to_print.size} hidden rows..)".center(len(header))
    end_rows = _rows_to_strings(end_rows, column_widths)
    return header + "\n".join(start_rows) + "\n" + cutoff_line + "\n" + "\n".join(end_rows)


def _rows_to_strings(rows: "FancyArray", column_widths: list[tuple[str, int]]) -> list[str]:
    rows_as_strings = []
    for row in rows.data:
        row_as_strings = []
        for attr, (_, width) in zip(row.tolist(), column_widths):
            row_as_strings.append(_center_and_truncate(str(attr), width))
        rows_as_strings.append("|".join(row_as_strings))
    return rows_as_strings


def _determine_column_widths(array: "FancyArray") -> list[tuple[str, int]]:
    """Get the maximum width of each column in the array."""
    column_widths: list[tuple[str, int]] = []
    if not array.dtype.names:
        return column_widths

    for column in array.dtype.names:
        data = array.data[column]
        if data.size:
            # if float, round to 3 decimals
            if data.dtype.kind == "f":
                data = np.around(data, decimals=3)
            # to string to get the length
            data = data.astype(str)
            longest_string = max(data, key=len)
            if len(column) > len(longest_string):
                longest_string = column
        else:
            longest_string = column
        column_widths.append((column, len(longest_string) + 2))
    return column_widths


def _center_and_truncate(string: str, width: int) -> str:
    if len(string) <= width:
        return string.center(width)
    return f"{string[: width - 2]}..".center(width)


def _get_start_and_end_rows(array: "FancyArray", rows: int) -> tuple["FancyArray", Optional["FancyArray"]]:
    if array.size <= rows:
        return array, None
    cutoff = rows // 2
    start_rows = array[:cutoff]
    end_rows = array[-cutoff:]
    return start_rows, end_rows



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.model.arrays import (
    AsymCurrentSensorArray,
    AsymLineArray,
    AsymVoltageSensorArray,
    Branch3Array,
    BranchArray,
    GenericBranchArray,
    IdArray,
    LineArray,
    LinkArray,
    NodeArray,
    SourceArray,
    SymCurrentSensorArray,
    SymGenArray,
    SymLoadArray,
    SymPowerSensorArray,
    SymVoltageSensorArray,
    ThreeWindingTransformerArray,
    TransformerArray,
    TransformerTapRegulatorArray,
)

__all__ = [
    "IdArray",
    "NodeArray",
    "BranchArray",
    "LinkArray",
    "LineArray",
    "GenericBranchArray",
    "AsymLineArray",
    "TransformerArray",
    "Branch3Array",
    "ThreeWindingTransformerArray",
    "SourceArray",
    "SymGenArray",
    "SymLoadArray",
    "TransformerTapRegulatorArray",
    "AsymVoltageSensorArray",
    "AsymCurrentSensorArray",
    "SymPowerSensorArray",
    "SymVoltageSensorArray",
    "SymCurrentSensorArray",
]
